function [allBlurKernels4D ell2norm2_map]=makeFovBlurKernels(ell1norm,targetEll2norm2,rho,theta,extraPad,patchOffset)
%
% Builds the constrained set of blur kernels for the foveation operator:       (ver 2.12, May 15, 2016)
% each BlurKernel has ell^1 norm equal to ell1norm and the PFS to be used when blurring
% pixels at coordinates (i,j) has ell^2 norm equal to targetEll2norm2(i,j)
%
%
%  REFERENCES:
% [1] A. Foi and G. Boracchi, "Foveated self-similarity in nonlocal image filtering",
%     Proc. IS&T/SPIE EI2012 - Human Vision and Electronic Imaging XVII, 8291-32,
%     Burlingame (CA), USA, Jan. 2012.  http://doi.org/10.1117/12.912217
% [2] A. Foi and G. Boracchi, "Anisotropically Foveated Nonlocal Image Denoising",
%     Proc. IEEE Int. Conf. Image Process. (ICIP 2013), pp. 464-468, Melbourne, Australia, Sep. 2013.
%     http://doi.org/10.1109/ICIP.2013.6738096
% [3] A. Foi and G. Boracchi, "Foveated nonlocal self-similarity", to appear in Int. J. Computer Vision, 2016.
%     http://doi.org/10.1007/s11263-016-0898-1
% 
%  All available online at http://www.cs.tut.fi/~foi/
%
%
%  author:  Alessandro Foi, Tampere University of Technology, Finland
%           All rights reserved.
% --------------------------------------------------------------------------------------
%


if ~exist('extraPad','var')||isempty(extraPad);
    extraPad=0;
end
if ~exist('rho','var')||isempty(rho);
    rho=1;
end
if ~exist('theta','var')||isempty(theta);
    theta=0;
end
if ~exist('patchOffset','var')||isempty(patchOffset);
    patchOffset=[0 0];
end

[size_U_1,size_U_2]=size(targetEll2norm2); % size of the ell-2 norm map; coincides with the size of k_kernel and size of the foveated patch.

% upper and lower bounds for the squared l2norm of the blur kernel (needed to guarantee foveation operator properties)
ell2norm_Ubound=ell1norm^2;   % this is attained for a Dirac-delta impulse
ell2norm_Lbound=ell1norm^2/((size_U_2+2*extraPad)*(size_U_1+2*extraPad));  % this is attained for a uniform BlurKernel over the whole blur kernel support.
targetEll2norm2=min(ell2norm_Ubound-sqrt(eps),max(targetEll2norm2,ell2norm_Lbound+sqrt(eps)));

ell2norm2_map=zeros(size_U_1,size_U_2);
allBlurKernels4D =zeros(size_U_1+extraPad*2,size_U_2+extraPad*2,size_U_1,size_U_2);


if rho~=1
    textString=['anisotropic foveation: rho=',num2str(rho), ', theta=',num2str(theta)];
else
    textString='isotropic foveation';
end
textString=[textString, ', patch size=',num2str(size_U_2),'x',num2str(size_U_1)];

varsigmaInit=ell1norm./(2*sqrt(pi*targetEll2norm2));
if any(extraPad)
    disp(' - Constructing blur kernels to be used by the foveation operator ...  ')
    textString=[textString,', extraPad=',num2str(extraPad)];
else
    disp(' - Constructing "patch-supported" blur kernels to be used by the foveation operator ...  ')
    textString=[textString,'  | self-map'];
end

if any(patchOffset);
    textString=[textString,'  patchOffset=[',num2str(patchOffset(1)),',',num2str(patchOffset(2)),']'];
end
disp([' . ',textString]);

timeFirst=now;
timeOld=timeFirst;
loopCount=0;
loopCountPartial=0;
loopTotal=size_U_1*size_U_2;
progressStringEveryNseconds=2.0;
timeLoop=inf;


for u_1=1:size_U_1;
    for u_2=1:size_U_2;
        
        % define cost function fitCostFun  which penalizes the difference between ell2norm2 of the blur kernel and targetEll2norm2.
        fitCostFun=@(ell2norm2,targetEll2norm2) abs(ell2norm2-targetEll2norm2);
        fTol=targetEll2norm2(u_1,u_2)*eps('single'); % numerical tolerance of fitCostFun to decide convergence
        MaxFunEvals=40; % maximum number of evaluations of the cost function
        showFig=0; % show blur kernel figure during optimization?
        varsigma=varsigmaInit(u_1,u_2); % value of varsigma used as initialization for the iterative optimization
        
        if 1 % use iterative scaling of varsigma
            fitError=+inf;
            funEvalCount=0;
            while fitError>fTol
                funEvalCount=funEvalCount+1;
                ell2norm2=blurEll2norm2(u_1,u_2,size_U_1,size_U_2,extraPad,varsigma,ell1norm,rho,theta,showFig,patchOffset);
                fitError=fitCostFun(ell2norm2,targetEll2norm2(u_1,u_2));
                if fitError>fTol
                    ratioNew=sqrt(ell2norm2/targetEll2norm2(u_1,u_2)); % varsigma is (very roughly) inversely proportional to sqrt(ell2norm2), therefore this is a good guess of the scaling factor for correcting varsigma
                    if funEvalCount>1
                        ratioPow=min(10,max(0.5,ratioPow*log(ratioOld)/log(ratioOld/ratioNew)));  % here we try to guess how many times one should actually have to multiply varsigma by ratioNew to get the ratio sqrt(ell2norm2/targetEll2norm2(u_1,u_2)) to equal one. ratioPow is approximately 1 whenever the above inverse proportionality holds accurately. ratioPow becomes very large when targetEll2norm2 is close to the its upper limiting value, i.e. when the blur kernel should become a Dirac.  This provides a substantial acceleration of the convergence.
                    else
                        ratioPow=1;
                    end
                    varsigma=varsigma*ratioNew^ratioPow;
                    ratioOld=ratioNew;
                end
                if funEvalCount>=MaxFunEvals
                    break
                end
            end
        else % varsigma optimization using Nelder-Mead algorithm (fminsearch)
            MaxIter=ceil(MaxFunEvals/2);  % maximum number of iterations of fminsearch
            % optimSetDisplay='iter';
            optimSetDisplay='none';
            xTol=varsigma*eps('single');  % numerical tolerance of varsigma to decide convergence
            
            % use fminsearch to find correct value of varsigma (it is necessary to optimize varsigma if extraPad=0, since blur kernels may have a support that does not fit over the actual domain of the desired foveated patch)
            [varsigma fitError]=fminsearch(@(varsigma) fitCostFun(blurEll2norm2(u_1,u_2,size_U_1,size_U_2,extraPad,varsigma,ell1norm,rho,theta,showFig,patchOffset),targetEll2norm2(u_1,u_2)),varsigma,optimset('MaxFunEvals',MaxFunEvals,'MaxIter',MaxIter,'TolX',xTol,'TolFun',fTol,'Display',optimSetDisplay));
            
        end
        showFig=0;
        [ell2norm2 blurKernel]=blurEll2norm2(u_1,u_2,size_U_1,size_U_2,extraPad,varsigma,ell1norm,rho,theta,showFig,patchOffset);
        ell2norm2_map(u_1,u_2)=ell2norm2;
        allBlurKernels4D (:,:,u_1,u_2)=blurKernel;
        
        
        % progress update string
        loopCount=loopCount+1;
        loopCountPartial=loopCountPartial+1;
        if timeLoop*loopCountPartial>progressStringEveryNseconds;
            timeNow=now;
            timeLoop=86400*(timeNow-timeOld)/loopCountPartial;
            if 86400*(timeNow-timeOld)>progressStringEveryNseconds;
                timeOld=timeNow;
                loopCountPartial=0;
                disp(['      ',num2str(floor(100*loopCount/loopTotal)),' %  (ETA ',num2str(timeLoop*(loopTotal-loopCount)),'s)'])
            end
        end
    end
end
disp(['     ',num2str(100),' %  completed  (',num2str(86400*(now-timeFirst)),'s)'])


% % this is a figure to verify the accuracy of the optimization of varsigma:
% fitError=abs(ell2norm2_map-targetEll2norm2);  figure, imagesc(log10(fitError),[-13 -3]),colorbar;

end

%%


function [ell2norm2,blurKernel]=blurEll2norm2(u_1,u_2,size_U_1,size_U_2,extraPad,varsigma,ell1norm,rho,theta,showFig,patchOffset)
% this function builds a blur kernel with a given ell1norm and varsigma value
% OUTPUTS:
%   ell2norm2 squared \ell-2 norm of the blur kernel
%   blurKernel is the blur kernel
%
% INPUTS:
%   u_1, u_2   coordinate of the center of the blur kernel
%   size_U_1, size_U_2    size of the foveated patch
%   extraPad   extra padding of the patch domain
%   ell1norm   ell1norm of the blur kernel
%   rho,theta    anisotropy parameters
%   patchOffset  spatial shift of the operator (must be absent or set to [0,0])
%   showFig    useful to display blur kernels during optimization


if varsigma<=0  % minimal degenerate case
    
    blurKernel=zeros(2*extraPad+[size_U_1,size_U_2]);  blurKernel(u_1+extraPad,u_2+extraPad)=1;     % discrete Dirac
    
elseif varsigma==inf; % maximal degenerate case
    
    blurKernel=ones(2*extraPad+[size_U_1,size_U_2]);   % uniform BlurKernel
    
else
    
    % center of the patch
    Ucenter_1=(size_U_1+1)/2;
    Ucenter_2=(size_U_2+1)/2;
    
    if exist('patchOffset','var');
        u_1=u_1+patchOffset(1);
        u_2=u_2+patchOffset(2);
        Ucenter_1=Ucenter_1+patchOffset(1);
        Ucenter_2=Ucenter_2+patchOffset(2);
    end
    
    
    % oversampling factor
    oversamplingFactor=5;   % use oversampling to mitigate issues with discretization of Gaussian kernels
    oversamplingFactor=ceil((oversamplingFactor-1)/2)*2+1;   % oversamplingFactor must be an odd positive integer
    
    % create oversampled rotated grid with origin at the impulse position
    [grid_2,grid_1]=meshgrid((round((-extraPad+1-u_2)*oversamplingFactor-(oversamplingFactor-1)/2):round((size_U_2+extraPad-u_2)*oversamplingFactor+(oversamplingFactor-1)/2))/oversamplingFactor,(round((-extraPad+1-u_1)*oversamplingFactor-(oversamplingFactor-1)/2):round((size_U_1+extraPad-u_1)*oversamplingFactor+(oversamplingFactor-1)/2))/oversamplingFactor);
    
    theta_prime=atan2((Ucenter_2-u_2),(Ucenter_1-u_1))+theta;     % theta_prime gives the rotation between the cardinal axes of the blur kernel and the cardinal axes of the patch.
    gridRotTan=grid_2*cos(theta_prime)-grid_1*sin(theta_prime);   % this is tangential if theta=0;
    gridRotRad=grid_2*sin(theta_prime)+grid_1*cos(theta_prime);   % this is radial if theta=0;
    
    
    % define standard deviations sigmaTan and sigmaRad of the bivariate Gaussian blur kernel with respect to the rotated grid.
    radius2=(Ucenter_2-u_2)^2+(Ucenter_1-u_1)^2;
    if radius2==0
        rho=1;   % enforce the circularly-symmetric blur kernel at the fixation point
    elseif ~exist('rho','var')
        rho=4.0;   % default to anisotropic with rho=4.0
    end
    sigmaTan=varsigma/sqrt(rho);  % this is tangential if theta=0;
    sigmaRad=varsigma*sqrt(rho);  % this is radial if theta=0;
    
    % build un-normalized elliptical Gaussian blur kernel on oversampled rotated grid
    blurKernel= exp( -( gridRotTan.^2/(2*sigmaTan^2) + gridRotRad.^2/(2*sigmaRad^2) ) )  /  (2*pi*sigmaTan*sigmaRad);
    
    % downsample to remove oversampling
    blurKernel=conv2(blurKernel,ones(oversamplingFactor)/oversamplingFactor^2,'same');  % basic antialiasing using uniform kernel
    blurKernel=blurKernel((oversamplingFactor-1)/2+1:oversamplingFactor:end-(oversamplingFactor-1)/2,(oversamplingFactor-1)/2+1:oversamplingFactor:end-(oversamplingFactor-1)/2);  % decimation
end

blurKernel=blurKernel./sum(blurKernel(:))*ell1norm;  % enforce l1-norm of blurKernel  (in older codes, this used to be sqrt(-2*pi/log(1-p))*sqrt(kappa_0))
ell2norm2=sum(sum(blurKernel.*blurKernel));          % compute l2-norm^2 of blurKernel

if exist('showFig','var')&&showFig
    figure(5)
    axisHandle=gca;
    image(blurKernel,'parent',axisHandle,'CDataMapping','scaled');colormap(hot)
    title(axisHandle,['u=[',num2str(u_1),',',num2str(u_2),']  \varsigma=',num2str(varsigma),'  squared \ell-2 norm =',num2str(ell2norm2)]);
    colorbar,axis equal tight
    pause(eps)
end

end
